﻿Option Strict On
Option Infer On

Imports Helpers.Mathematics
Imports D2DControlset
Imports SlimDX.Direct2D
Imports SlimDX.DirectWrite
Imports Helpers.Direct2D

Friend MustInherit Class Entity
    Private WithEvents w As World

    Friend Property world As World
        Get
            Return w
        End Get
        Set(value As World)
            w = value
        End Set
    End Property

    Friend Property InUse As Boolean

    Friend MustOverride Function getLocation() As PointF

    Friend MustOverride Function getSize() As SizeF

    Friend MustOverride Sub doStep()

    Friend Overridable Sub draw(ByVal e As D2DDrawEventArgs, ByVal offset As PointF)

    End Sub

    Friend Overridable Sub onSpawned()

    End Sub

    Friend Overridable Sub onDespawned()

    End Sub

    Friend Overridable Function getRect() As RectangleF
        Return New RectangleF(getLocation.X - getSize.Width / 2.0F, getLocation.Y - getSize.Height / 2.0F, getSize.Width, getSize.Height)
    End Function
End Class

Friend Class Explosion
    Inherits Entity

    Private sound As SoundEffect
    Private m_Location As PointF
    Private particles As List(Of Particle)
    Private rotation As Single
    Private thorns As Integer
    Private steps As Integer
    Private size As Single
    'Private showEffect As Boolean
    'Private op As Single
    Private playSound As Boolean

    Friend Overrides Function getLocation() As PointF
        Return m_Location
    End Function

    Friend Overrides Function getSize() As SizeF
        Return New SizeF(size, size)
    End Function

    Friend Overrides Sub doStep()
        Dim toRemove As Integer = Math.Min(Game.rand.Next(4), particles.Count)
        For i = 1 To toRemove
            Dim pIndex As Integer = Game.rand.Next(particles.Count)
            Dim p As Particle = particles.Item(pIndex)
            world.despawnEntity(p)
            particles.Remove(p)
        Next
        If particles.Count <= 0 Then world.despawnEntity(Me)
        'Select Case steps
        '    Case Is <= 0
        '        steps += 1
        '        Exit Sub
        '    Case Is <= 15
        '        op = steps / 15.0F
        '    Case Is <= 75
        '        op = 1
        '    Case Is <= 105
        '        op = (105 - steps) / 30.0F
        '    Case Else
        '        op = 0.0F
        'End Select
        steps += 1
    End Sub

    Sub New(ByVal p As PointF, ByVal colors() As Color, ByVal mainCol As Color, ByVal count As Integer, ByVal size As Single, ByVal showEffect As Boolean, ByVal playSound As Boolean)
        m_Location = p
        Me.size = size
        'Me.showEffect = showEffect
        Me.playSound = playSound
        particles = New List(Of Particle)
        rotation = Convert.ToSingle(Game.rand.NextDouble * Math.PI)
        thorns = Game.rand.Next(6, 11)
        Dim weights(colors.Length - 1) As Integer
        For i = 0 To colors.Length - 1
            If colors(i).ToArgb = mainCol.ToArgb Then
                weights(i) = 100
            Else
                weights(i) = 10
            End If
        Next
        Dim indices(colors.Length - 1) As Integer
        For i = 0 To colors.Length - 1
            indices(i) = i
        Next
        For i = 1 To count
            Dim cIndex As Integer = MathHelper.getRandomNumber(indices, weights)
            Dim c As Color = colors(cIndex)
            Dim v As PointF = MathHelper.getPointOnCircle(New Helpers.Mathematics.Vector2, Game.rand.NextDouble * 2 * Math.PI, Game.rand.NextDouble * 3 + 2).toPointF
            Dim part As New Particle(v, c, m_Location, New SizeF(1.5F, 1), True)
            particles.Add(part)
        Next
    End Sub

    Friend Overrides Sub onSpawned()
        MyBase.onSpawned()
        If playSound Then
            sound = New SoundEffect("c01.wma", True)
            sound.play()
        End If
        For i = 0 To particles.Count - 1
            world.spawnEntity(particles.Item(i))
        Next
    End Sub

    Friend Overrides Sub draw(e As D2DDrawEventArgs, offset As PointF)
        'If Not showEffect Then Exit Sub
        'If op <= 0 Then Exit Sub
        'Dim r As Single = op * size
        'Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        'Dim mid As New PointF(offset.X + getLocation.X, offset.Y + getLocation.Y)
        'Dim b1 As New SolidColorBrush(e.Target, New SlimDX.Color4(Color.OrangeRed))
        'Dim b2 As New SolidColorBrush(e.Target, New SlimDX.Color4(Color.Yellow))
        'b1.Opacity = op * 0.5F
        'b2.Opacity = op * 0.5F
        'h.FillStar(mid, r, rotation, b1, thorns)
        'h.FillStar(mid, r * 0.6F, rotation, b2, thorns)
        'b2.Dispose()
        'b1.Dispose()
    End Sub
End Class

Friend Class Particle
    Inherits Entity

    Protected moveVec As PointF
    Private col As Color
    Protected m_Location As PointF
    Protected m_Size As SizeF
    Private m_reflect As Boolean

    Friend ReadOnly Property color As Color
        Get
            Return col
        End Get
    End Property

    Friend Overrides Function getLocation() As PointF
        Return m_Location
    End Function

    Friend Overrides Function getSize() As SizeF
        Return m_Size
    End Function

    Friend Overrides Sub doStep()
        m_Location.X += moveVec.X
        m_Location.Y += moveVec.Y
        If world Is Nothing Then Exit Sub
        If m_Location.X <= 0 OrElse m_Location.X >= world.size.Width Then
            If m_reflect Then
                moveVec = New PointF(-moveVec.X, moveVec.Y)
            Else
                world.despawnEntity(Me)
                Exit Sub
            End If
        End If
        If m_Location.Y <= 0 OrElse m_Location.Y >= world.size.Height Then
            If m_reflect Then
                moveVec = New PointF(moveVec.X, -moveVec.Y)
            Else
                world.despawnEntity(Me)
                Exit Sub
            End If
        End If
    End Sub

    Sub New(ByVal v As PointF, ByVal c As Color, ByVal loc As PointF, ByVal s As SizeF, ByVal reflect As Boolean)
        moveVec = v
        col = c
        m_Location = loc
        m_Size = s
        m_reflect = reflect
    End Sub

    Friend Overrides Sub draw(e As D2DDrawEventArgs, offset As PointF)
        Dim mv As New Vector2(getLocation.X + offset.X, getLocation.Y + offset.Y)
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim b As New SolidColorBrush(e.Target, New SlimDX.Color4(color))
        b.Opacity = 0.7F
        h.DrawCircle(mv, m_Size.Width, b, m_Size.Height)
        b.Dispose()
    End Sub
End Class