﻿Imports SlimDX
Imports SlimDX.Direct2D
Imports SlimDX.DirectWrite
Imports D2DControlset

Friend Class TDTextBox : Inherits GameControl
    Private selected As Boolean
    Private cursorIndex As Integer
    Private cursorOpacity As Integer
    Private cursorEffect As Boolean
    Private lines As List(Of String)
    Private currentLine As Integer

    Friend Shadows Property Text As String
        Get
            Return String.Join(Environment.NewLine, lines)
        End Get
        Set(value As String)
            Dim l() = value.Split({Environment.NewLine}, StringSplitOptions.None)
            lines.Clear()
            lines.AddRange(l)
            If lines.Count = 0 Then lines.Add(String.Empty)
            currentLine = lines.Count - 1
            cursorIndex = lines.Last.Length
        End Set
    End Property

    Friend Property MultiLine As Boolean

    Friend Property IsNumeric As Boolean

    Friend Property IsReadOnly As Boolean

    Protected Overrides Sub drawContent(e As D2DDrawEventArgs, inRect As RectangleF)
        Dim textRect = inRect
        Dim layout As TextLayout = e.TextFactory.CreateTextLayout(Text, FontManager.GetFont("defaultFormat"), textRect.Width, textRect.Height)
        e.Target.DrawTextLayout(New PointF(textRect.X, textRect.Y), layout, ColorManager.GetBrush("defaultWhite"))
        layout.Dispose()
        If Not selected Then Exit Sub
        If cursorEffect Then
            cursorOpacity -= 3
            If cursorOpacity <= 0 Then cursorEffect = False
        Else
            cursorOpacity += 3
            If cursorOpacity >= 100 Then cursorEffect = True
        End If
        Dim s1 = lines(currentLine).Substring(0, cursorIndex)
        Dim s = Game.window.measureString(s1, FontManager.GetFont("defaultFormat"))
        Dim b As New SolidColorBrush(e.Target, CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush).Color)
        b.Opacity = cursorOpacity / 100.0F
        e.Target.FillRectangle(b, getCursorRect(textRect))
        b.Dispose()
    End Sub

    Private Function getCursorRect(textRect As RectangleF) As RectangleF
        Dim lineIndex = 0
        Dim linesBefore = 0
        Dim realLines = getRealLines(lines(currentLine), textRect)
        Dim charsLeft = cursorIndex - realLines.Count + 1
        If MultiLine Then
            Do Until charsLeft <= realLines(lineIndex).Length
                charsLeft -= realLines(lineIndex).Length
                lineIndex += 1
            Loop
            For i = 0 To currentLine - 1
                linesBefore += getRealLines(lines(i), textRect).Count
            Next
        End If
        Return New RectangleF(textRect.X + getWidth(realLines(lineIndex).Substring(0, charsLeft)), textRect.Y + (lineIndex + linesBefore) * 32 * Game.getSizeFactor, 2 * Game.getSizeFactor, 30 * Game.getSizeFactor)
    End Function

    Private Function getRealLines(line As String, textRect As RectangleF) As List(Of String)
        Dim realLines As New List(Of String)
        Dim words = line.Split(" "c)
        Dim i = 0
        Dim nextWord = words(i)
        Do While i < words.Length
            Dim currentWidth = getWidth(nextWord)
            Dim currentLine = String.Empty
            Do While currentWidth < textRect.Width AndAlso i < words.Length
                currentLine = String.Concat(currentLine, nextWord)
                i += 1
                If i < words.Length Then nextWord = String.Concat(" ", words(i))
                currentWidth += getWidth(nextWord)
            Loop
            realLines.Add(currentLine)
            If i < words.Length Then nextWord = words(i)
        Loop
        Return realLines
    End Function

    Private Function getWidth(s As String) As Single
        Return Game.window.measureString(s, FontManager.GetFont("defaultFormat")).Width
    End Function

    Sub New(parent As IDirect2DControl)
        MyBase.New(parent)
        lines = New List(Of String)
        lines.Add(String.Empty)
        currentLine = 0
        cursorIndex = 0
        setBorderWidth(2.0F)
        selected = False
        AddHandler m_Window.KeyPress, AddressOf KeyPress
        AddHandler m_Window.KeyDown, AddressOf KeyDown
        AddHandler m_Window.MouseDown, AddressOf WindowMouseDown
    End Sub

    Private Sub KeyPress(sender As Object, e As KeyPressEventArgs)
        If Not selected Then Exit Sub
        Dim s1 = lines(currentLine).Substring(0, cursorIndex)
        Dim s2 = If(cursorIndex >= lines(currentLine).Length, String.Empty, lines(currentLine).Substring(cursorIndex, lines(currentLine).Length - cursorIndex))
        If e.KeyChar = Convert.ToChar(Keys.Back) Then
            If cursorIndex > 0 Then
                lines(currentLine) = String.Concat(s1.Substring(0, s1.Length - 1), s2)
                cursorIndex -= 1
            Else
                If currentLine > 0 Then
                    cursorIndex = lines(currentLine - 1).Length
                    lines(currentLine - 1) &= lines(currentLine)
                    lines.RemoveAt(currentLine)
                    currentLine -= 1
                End If
            End If
        Else
            If Char.IsDigit(e.KeyChar) OrElse e.KeyChar = Convert.ToChar(".") OrElse e.KeyChar = Convert.ToChar(",") OrElse (Not IsNumeric AndAlso (Char.IsLetter(e.KeyChar) OrElse e.KeyChar = Convert.ToChar(Keys.Space))) Then
                lines(currentLine) = String.Concat(s1, e.KeyChar, s2)
                cursorIndex += 1
            Else
                If e.KeyChar = Convert.ToChar(Keys.Enter) AndAlso MultiLine Then
                    If cursorIndex < lines(currentLine).Length Then lines(currentLine) = lines(currentLine).Remove(cursorIndex)
                    lines.Add(s2)
                    currentLine += 1
                    cursorIndex = 0
                End If
            End If
        End If
    End Sub

    Private Sub KeyDown(sender As Object, e As KeyEventArgs)
        If Not selected Then Exit Sub
        If e.KeyCode = Keys.Left Then
            If cursorIndex > 0 Then
                cursorIndex -= 1
            Else
                If currentLine > 0 Then
                    currentLine -= 1
                    cursorIndex = lines(currentLine).Length
                End If
            End If
        End If
        If e.KeyCode = Keys.Right Then
            If cursorIndex < lines(currentLine).Length Then
                cursorIndex += 1
            Else
                If currentLine < lines.Count - 1 Then
                    currentLine += 1
                    cursorIndex = 0
                End If
            End If
        End If
    End Sub

    Private Sub WindowMouseDown(sender As Object, e As MouseEventArgs)
        Dim testRect As New Rectangle(AbsoluteLocation, Size)
        selected = Not IsReadOnly AndAlso testRect.Contains(e.Location)
    End Sub
End Class
