﻿Option Strict On
Option Infer On

Imports D2DControlset
Imports SlimDX
Imports SlimDX.Direct2D
Imports SlimDX.DirectWrite
Imports MessageBoxExtended
Imports Helpers.Direct2D
Imports System.Reflection
Imports System.Threading

Friend Class WorldControl : Inherits Menu
    Private WithEvents world As World
    Private offset As Point
    Private dragging As Boolean
    Private lastMousePoint As Point
    Private largometer As Lagometer
    Private map As Minimap
    Private iPic As InfoPicture
    Private iMenu As InfoMenu
    Private menuButton As IngameButton
    Private buildButton As IngameButton
    Private buildMenu As Buildmenu
    Private m_selectedTower As Tower
    Private selectSound As SoundEffect
    Private infoSidebar As InfoSidebar
    Private menu As IngameMenu

    Friend Property selectedTower As Tower
        Get
            Return m_selectedTower
        End Get
        Private Set(value As Tower)
            m_selectedTower = value
            refreshInfoMenu()
        End Set
    End Property

    Friend ReadOnly Property isBuilding As Boolean
        Get
            Return buildMenu.selectedItem IsNot Nothing
        End Get
    End Property

    Friend Function getOffset() As Point
        Return offset
    End Function

    Private Sub refreshInfoMenu()
        Dim setVisible = selectedTower IsNot Nothing
        iPic.Visible = setVisible
        iMenu.Visible = setVisible
        If setVisible Then
            iPic.image = m_selectedTower.getImage
            iMenu.towerName = m_selectedTower.name
            iMenu.towerLevel = m_selectedTower.level
            iMenu.towerRadius = Convert.ToInt32(m_selectedTower.radius)
            iMenu.towerDamage = m_selectedTower.damage
        End If
    End Sub

    Protected Overrides Sub OnDraw(e As D2DDrawEventArgs)
        Dim o = offset
        o.X += AbsoluteLocation.X
        o.Y += AbsoluteLocation.Y
        Dim r As New RectangleF(o.X - 0.5F, o.Y - 0.5F, world.size.Width + 1, world.size.Height + 1)
        e.Target.PushAxisAlignedClip(r, AntialiasMode.PerPrimitive)
        drawWay(e)
        drawSelected(e)
        drawEntities(e)
        drawBuildSelection(e)
        drawMapBorder(e)
        e.Target.PopAxisAlignedClip()
        If world.GamePaused Then
            Dim b As New SolidColorBrush(e.Target, New Color4(0, 0, 0))
            b.Opacity = 0.7
            e.Target.FillRectangle(b, ClientRectangle)
            b.Dispose()
        End If
        If world.MessageVisible Then DrawMessage(e)
    End Sub

    Private Sub DrawMessage(e As D2DDrawEventArgs)
        Dim formatName = "messageFormat"
        If Not FontManager.FontRegistered(formatName) Then FontManager.RegisterFont(formatName, New FontTemplate("Polentical Neon", 16, TextAlignment:=TextAlignment.Center, paragraphAlignment:=ParagraphAlignment.Center))
        Dim rectWidth = Convert.ToInt32(Width * 0.6)
        Dim rectHeight = Convert.ToInt32(150 * Game.getSizeFactor)
        Dim rect As New Rectangle((Width - rectWidth) \ 2 + AbsoluteLocation.X, Convert.ToInt32(100 * Game.getHeightFactor) + AbsoluteLocation.Y, rectWidth, rectHeight)
        e.Target.FillRectangle(ColorManager.GetBrush("defaultBlack"), rect)
        e.Target.DrawRectangle(ColorManager.GetBrush("defaultWhite"), rect, 3 * Game.getSizeFactor)
        Dim messageHeight = Convert.ToInt32(rectHeight * 0.7)
        Dim continueHeight = rectHeight - messageHeight
        Dim messageRect As New Rectangle(rect.X, rect.Y, rectWidth, messageHeight)
        Dim continueRect As New Rectangle(rect.X, rect.Y + messageHeight, rectWidth, continueHeight)
        e.Target.DrawText(world.MessageText, FontManager.GetFont(formatName), messageRect, ColorManager.GetBrush("defaultWhite"))
        Dim brushName = "continueBrush"
        If Not ColorManager.BrushRegistered(brushName) Then ColorManager.RegisterBrush(brushName, New SolidColorBrush(e.Target, New Color4(Color.Gold)))
        Dim format2Name = "continueFormat"
        If Not FontManager.FontRegistered(format2Name) Then FontManager.RegisterFont(format2Name, New FontTemplate("Polentical Neon", 12, TextAlignment:=TextAlignment.Center, paragraphAlignment:=ParagraphAlignment.Center))
        e.Target.DrawText(Game.guiINI.getValue("message_continue"), FontManager.GetFont(format2Name), continueRect, ColorManager.GetBrush(brushName))
    End Sub

    Private Sub drawWay(e As D2DDrawEventArgs)
        Dim b As New SolidColorBrush(e.Target, New Color4(Color.DimGray))
        b.Opacity = 0.3F
        Dim pathPoints(world.way.segments.Length - 1) As PointF
        For i = 0 To world.way.segments.Length - 1
            Dim seg = world.way.segments(i)
            pathPoints(i) = New PointF(seg.p2.X + offset.X + AbsoluteLocation.X, seg.p2.Y + offset.Y + AbsoluteLocation.Y)
        Next
        Dim firstP = world.way.segments.First.p1
        firstP.X += offset.X + AbsoluteLocation.X
        firstP.Y += offset.Y + AbsoluteLocation.Y
        Dim sProps As New StrokeStyleProperties
        sProps.DashCap = CapStyle.Round
        sProps.DashStyle = DashStyle.Solid
        sProps.EndCap = CapStyle.Flat
        sProps.StartCap = CapStyle.Round
        sProps.LineJoin = LineJoin.Round
        Dim style As New StrokeStyle(e.Factory, sProps)
        Dim path As New PathGeometry(e.Factory)
        Dim sink = path.Open
        sink.BeginFigure(firstP, FigureBegin.Hollow)
        sink.AddLines(pathPoints)
        sink.EndFigure(FigureEnd.Open)
        sink.Close()
        Dim endCircle As New Ellipse
        endCircle.Center = pathPoints.Last
        endCircle.RadiusX = 10
        endCircle.RadiusY = 10
        Dim eg As New EllipseGeometry(e.Factory, endCircle)
        Dim g As New PathGeometry(e.Factory)
        Dim s = g.Open
        Geometry.Combine(path, eg, CombineMode.Union, s)
        s.Close()
        e.Target.DrawGeometry(g, b, 24.0F, style)
        s.Dispose()
        g.Dispose()
        eg.Dispose()
        sink.Dispose()
        path.Dispose()
        style.Dispose()
        b.Dispose()
        For i = 0 To world.way.segments.Length - 1
            world.way.segments(i).Draw(e, offset)
        Next
    End Sub

    Private Sub drawMapBorder(e As D2DDrawEventArgs)
        Dim sProps As New StrokeStyleProperties
        sProps.DashStyle = DashStyle.Custom
        Dim style As New StrokeStyle(e.Factory, sProps, {5, 8})
        e.Target.DrawRectangle(ColorManager.GetBrush("defaultWhite"), New RectangleF(offset.X + AbsoluteLocation.X - 0.5F, offset.Y + AbsoluteLocation.Y - 0.5F, world.size.Width + 1, world.size.Height + 1), 1.0F, style)
        style.Dispose()
    End Sub

    Private Sub drawEntities(e As D2DDrawEventArgs)
        Dim o = offset
        o.X += AbsoluteLocation.X
        o.Y += AbsoluteLocation.Y
        Dim ents() = world.getEntities
        For i = 0 To ents.Length - 1
            Do While ents(i).InUse
                Thread.Sleep(1)
            Loop
            ents(i).InUse = True
            ents(i).draw(e, o)
            ents(i).InUse = False
        Next
    End Sub

    Private Sub drawBuildSelection(e As D2DDrawEventArgs)
        If buildMenu.selectedItem Is Nothing Then Exit Sub
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim mousePos As Point = PointToClient(Cursor.Position)
        Dim rect As New RectangleF(mousePos.X - 12 - offset.X, mousePos.Y - 12 - offset.Y, 24, 24)
        Dim b As SolidColorBrush
        Dim able As New SolidColorBrush(e.Target, New Color4(0, 1, 0.2F))
        Dim unable As New SolidColorBrush(e.Target, New Color4(0.7F, 0.1F, 0))
        If world.money >= buildMenu.selectedItem.costs AndAlso world.canBuild(rect) Then
            b = able
        Else
            b = unable
        End If
        b.Opacity = 0.5F
        Dim m As New Helpers.Mathematics.Vector2(mousePos.X + AbsoluteLocation.X, mousePos.Y + AbsoluteLocation.Y)
        h.FillCircle(m, 14, b)
        unable.Dispose()
        able.Dispose()
        h.DrawCircle(m, Convert.ToSingle(buildMenu.selectedItem.radius), ColorManager.GetBrush("defaultWhite"), 2)
    End Sub

    Private Sub drawSelected(e As D2DDrawEventArgs)
        If m_selectedTower Is Nothing Then Return
        Do While m_selectedTower.InUse
            Thread.Sleep(1)
        Loop
        m_selectedTower.InUse = True
        Dim h As New Direct2DHelper(e.Factory, e.Target, e.TextFactory)
        Dim m As New Helpers.Mathematics.Vector2(m_selectedTower.getLocation.X + AbsoluteLocation.X + offset.X, m_selectedTower.getLocation.Y + AbsoluteLocation.Y + offset.Y)
        Dim props As New RadialGradientBrushProperties
        props.CenterPoint = m.toPointF
        props.HorizontalRadius = Convert.ToSingle(m_selectedTower.radius)
        props.VerticalRadius = props.HorizontalRadius
        Dim s1 As New GradientStop
        s1.Color = New Color4(0.15F, 0.9F, 0.9F, 0.9F)
        s1.Position = 1
        Dim s2 As New GradientStop
        s2.Color = New Color4(0.05F, 0.9F, 0.9F, 0.9F)
        s2.Position = 0.7F
        Dim s3 As New GradientStop
        s3.Color = New Color4(0.02F, 0.9F, 0.9F, 0.9F)
        s3.Position = 0
        Dim stops As New GradientStopCollection(e.Target, {s1, s2, s3})
        Dim areaB As New RadialGradientBrush(e.Target, stops, props)
        Dim el As New Ellipse
        el.Center = props.CenterPoint
        el.RadiusX = props.HorizontalRadius
        el.RadiusY = props.VerticalRadius
        e.Target.FillEllipse(areaB, el)
        e.Target.DrawEllipse(ColorManager.GetBrush("defaultWhite"), el, 1.5F)
        areaB.Dispose()
        stops.Dispose()
        Dim r As Single = 30.0F
        Dim orange As New SolidColorBrush(e.Target, New Color4(Color.Orange))
        orange.Opacity = 0.4F
        h.FillCircle(m, r, orange)
        orange.Opacity = 0.8F
        h.DrawCircle(m, r, orange, 2.0F)
        orange.Dispose()
        If m_selectedTower.getTarget Is Nothing Then
            m_selectedTower.InUse = False
            Return
        End If
        Dim red As New SolidColorBrush(e.Target, New Color4(Color.Red))
        red.Opacity = 0.4F
        Dim mt As New Helpers.Mathematics.Vector2(m_selectedTower.getTarget.getLocation.X + AbsoluteLocation.X + offset.X, m_selectedTower.getTarget.getLocation.Y + AbsoluteLocation.Y + offset.Y)
        Dim rt As Single = 20.0F
        h.FillCircle(mt, rt, red)
        red.Opacity = 0.8F
        h.DrawCircle(mt, rt, red, 2.0F)
        Dim sProps As New StrokeStyleProperties
        sProps.DashStyle = DashStyle.Dash
        Dim style As New StrokeStyle(e.Factory, sProps)
        e.Target.DrawLine(red, m.toPointF, mt.toPointF, 4.0F, style)
        style.Dispose()
        red.Dispose()
        m_selectedTower.InUse = False
    End Sub

    Friend Sub start()
        world.start(10, 0, Me)
    End Sub

    Private Sub Window_KeyDown(sender As Object, e As KeyEventArgs)
        If world.MessageVisible Then world.HideMessage()
        If world.GamePaused Then Return
        If e.KeyCode = Keys.F1 Then largometer.Visible = Not largometer.Visible
        If e.KeyCode = Keys.Escape Then ShowMenu()
    End Sub

    Private Sub DisableAll()
        map.Visible = False
        menuButton.Visible = False
        buildButton.Visible = False
        infoSidebar.Visible = False
        iPic.Visible = False
        iMenu.Visible = False
        buildMenu.Visible = False
        largometer.Visible = False
    End Sub

    Sub New(w As World)
        world = w
        AddHandler world.Won, Sub()
                                  DisableAll()
                                  Dim m As New WonLostMenu(Me, Game.guiINI.getValue("WinMessage"))
                                  m.Size = Size
                                  AddHandler m.LeaveGame, Sub()
                                                              world.stop()
                                                              goUp()
                                                          End Sub
                              End Sub
        AddHandler world.Lost, Sub()
                                   DisableAll()
                                   Dim m As New WonLostMenu(Me, Game.guiINI.getValue("DefeatMessage"))
                                   m.Size = Size
                                   AddHandler m.LeaveGame, Sub()
                                                               world.stop()
                                                               goUp()
                                                           End Sub
                               End Sub

        map = New Minimap(world)
        map.D2DParent = Me

        iPic = New InfoPicture
        iPic.D2DParent = Me
        iPic.Visible = False

        iMenu = New InfoMenu
        iMenu.D2DParent = Me
        iMenu.Visible = False
        AddHandler iMenu.upgradeRequested, Sub()
                                               If selectedTower IsNot Nothing AndAlso Not world.GamePaused AndAlso Not world.MessageVisible Then
                                                   selectedTower.levelUp()
                                                   refreshInfoMenu()
                                               End If
                                           End Sub
        AddHandler iMenu.upgradeCostShow, Sub() If selectedTower IsNot Nothing AndAlso buildMenu.selectedItem Is Nothing AndAlso Not world.GamePaused AndAlso Not world.MessageVisible Then infoSidebar.toPay = selectedTower.getLevelCost
        AddHandler iMenu.upgradeCostHide, Sub() If buildMenu.selectedItem Is Nothing AndAlso Not world.GamePaused AndAlso Not world.MessageVisible Then infoSidebar.toPay = 0

        menuButton = New IngameButton
        menuButton.D2DParent = Me
        menuButton.Text = "Menü [Esc]"
        AddHandler menuButton.Click, Sub() If Not world.GamePaused AndAlso Not world.MessageVisible Then ShowMenu()

        buildButton = New IngameButton
        buildButton.D2DParent = Me
        buildButton.Text = "Bauen [B]"
        AddHandler buildButton.Click, Sub() If Not world.GamePaused AndAlso Not world.MessageVisible Then buildMenu.Visible = Not buildMenu.Visible

        infoSidebar = New InfoSidebar
        infoSidebar.D2DParent = Me
        infoSidebar.waveInfo = "Welle 0 von " & world.waves

        Dim towers = Tower.getAllTowers()
        Dim genericClass = GetType(BuildmenuEntry(Of ))
        Dim entries As New List(Of BuildmenuEntry)
        Dim flags = BindingFlags.Static Or BindingFlags.NonPublic
        For i = 0 To towers.Length - 1
            If world.IsTowerAllowed(i) Then
                Dim t = towers(i)
                Dim constructed = genericClass.MakeGenericType(t)
                Dim name = DirectCast(t.GetMethod("getName", flags).Invoke(Nothing, {}), String)
                Dim costs = DirectCast(t.GetMethod("getCosts", flags).Invoke(Nothing, {}), Integer)
                Dim radius = DirectCast(t.GetMethod("getRadius", flags).Invoke(Nothing, {}), Integer)
                Dim entry = DirectCast(Activator.CreateInstance(constructed, {name, costs, Game.towerImages(i), radius}), BuildmenuEntry)
                entries.Add(entry)
            End If
        Next
        buildMenu = New Buildmenu(entries)
        buildMenu.D2DParent = Me
        buildMenu.Visible = False
        AddHandler buildMenu.selectedItemChanged, Sub()
                                                      If world.GamePaused OrElse world.MessageVisible Then Return
                                                      If buildMenu.selectedItem Is Nothing Then
                                                          infoSidebar.toPay = 0
                                                      Else
                                                          infoSidebar.toPay = buildMenu.selectedItem.costs
                                                          selectedTower = Nothing
                                                      End If
                                                  End Sub
        largometer = New Lagometer
        largometer.D2DParent = Me
        largometer.Visible = False

        menu = New IngameMenu(Me)
        menu.Visible = False
        AddHandler menu.LeaveGame, Sub()
                                       world.stop()
                                       goUp()
                                   End Sub
        AddHandler menu.ResumeGame, Sub()
                                        menu.Visible = False
                                        world.resume()
                                    End Sub

        selectSound = New SoundEffect("bt02.wma")
        sizeControls()
        setOffset()
        AddHandler Game.window.KeyDown, AddressOf Window_KeyDown
    End Sub

    Private Sub ShowMenu()
        world.pause()
        dragging = False
        menu.Visible = True
    End Sub

    Private Sub sizeControls()
        Dim f As Single = Game.getSizeFactor
        Dim lSize As Integer = Convert.ToInt32(300 * f)
        largometer.Size = New Size(lSize, lSize * 2)
        Dim mSize As Integer = Convert.ToInt32(200 * f)
        map.Size = New Size(mSize, mSize)
        Dim pSize As Integer = Convert.ToInt32(140 * f)
        iPic.Size = New Size(pSize, pSize)
        largometer.Location = New Point(Me.Width - largometer.Width, Me.Height - largometer.Height)
        map.Location = New Point(0, 0)
        iPic.Location = New Point(0, Me.Height - iPic.Height)
        iMenu.Size = New Size(Me.Width - (iPic.Width + iPic.X), Convert.ToInt32(80 * f))
        iMenu.Location = New Point(iPic.X + iPic.Width, Me.Height - iMenu.Height)
        Dim bWidth1 As Integer = Convert.ToInt32(110 * f)
        Dim bWidth2 As Integer = Convert.ToInt32(130 * f)
        Dim bHeight As Integer = Convert.ToInt32(35 * f)
        Dim bBorder As Integer = Convert.ToInt32(10 * f)
        buildButton.Size = New Size(bWidth1, bHeight)
        buildButton.Location = New Point(Width - bWidth1 - bBorder, bBorder)
        menuButton.Size = New Size(bWidth2, bHeight)
        menuButton.Location = New Point((Width - bWidth2) \ 2, bBorder)
        buildMenu.Location = New Point(buildButton.X, buildButton.Y + buildButton.Height + bBorder)
        buildMenu.Width = buildButton.Width
        buildMenu.Height = buildMenu.OptimalHeight
        infoSidebar.Location = New Point(0, map.Height + Convert.ToInt32(30 * f))
        infoSidebar.Size = New Size(map.Size.Width, map.Size.Height \ 2)
        menu.Size = Size
    End Sub

    Private Sub WorldControl_hide(sender As Object, e As EventArgs) Handles Me.hide
        Dim env As MusicEnvironment = MusicEnvironment.getEnvironment("menu")
        MusicManager.environment = env
        Me.D2DParent = Nothing
    End Sub

    Private Sub WorldControl_Click(sender As Object, e As MouseEventArgs) Handles Me.MouseClick
        If world.GamePaused OrElse world.MessageVisible Then Return
        Dim curPos As New Point(e.X - offset.X, e.Y - offset.Y)
        Dim ents() = world.getEntities
        If e.Button = MouseButtons.Right Then
            For Each ent In ents
                Do While ent.InUse
                    Thread.Sleep(1)
                Loop
                ent.InUse = True
                If TypeOf ent Is Creep AndAlso selectedTower IsNot Nothing AndAlso ent.getRect.Contains(curPos) Then
                    Dim c = DirectCast(ent, Creep)
                    If selectedTower.trySetTarget(c) Then
                        ent.InUse = False
                        Exit For
                    End If
                End If
                ent.InUse = False
            Next
        End If
        If e.Button = MouseButtons.Left Then
            If buildMenu.selectedItem Is Nothing Then
                Dim found = False
                For Each ent In ents
                    Do While ent.InUse
                        Thread.Sleep(1)
                    Loop
                    ent.InUse = True
                    If TypeOf ent Is Tower AndAlso ent.getRect.Contains(curPos) Then
                        selectedTower = DirectCast(ent, Tower)
                        found = True
                        selectSound.play()
                        ent.InUse = False
                        Exit For
                    End If
                    ent.InUse = False
                Next
                If Not found Then
                    If Not (buildButton.ClientRectangle.Contains(e.Location) _
                            OrElse (buildMenu.Visible AndAlso buildMenu.ClientRectangle.Contains(e.Location)) _
                            OrElse (iPic.Visible AndAlso iPic.ClientRectangle.Contains(e.Location)) _
                            OrElse (iMenu.Visible AndAlso iMenu.ClientRectangle.Contains(e.Location))) _
                        Then
                        If selectedTower IsNot Nothing Then
                            selectedTower = Nothing
                            selectSound.play()
                        End If
                    End If
                End If
                Exit Sub
            End If
            If Not buildMenu.Visible Then Return
            If buildMenu.ClientRectangle.Contains(e.Location) _
                OrElse buildButton.ClientRectangle.Contains(e.Location) _
                OrElse menuButton.ClientRectangle.Contains(e.Location) Then Return
            If buildMenu.selectedItem Is Nothing Then Return
            Dim t As Tower = buildMenu.selectedItem.getTower
            t.build(world, New Point(e.X - offset.X, e.Y - offset.Y))
        End If
    End Sub

    Private Sub WorldControl_MouseDown(sender As Object, e As MouseEventArgs) Handles Me.MouseDown
        If world.GamePaused OrElse world.MessageVisible Then Return
        If e.Button <> MouseButtons.Middle Then Return
        dragging = True
        lastMousePoint = e.Location
    End Sub

    Private Sub WorldControl_MouseMove(sender As Object, e As MouseEventArgs) Handles Me.MouseMove
        If world.GamePaused OrElse world.MessageVisible Then Return
        If Not dragging Then Return
        Dim moveX As Integer = e.X - lastMousePoint.X
        Dim moveY As Integer = e.Y - lastMousePoint.Y
        If world.size.Width > Size.Width Then
            offset.X += moveX
            If offset.X < Size.Width - world.size.Width Then offset.X = Size.Width - world.size.Width
            If offset.X > 0 Then offset.X = 0
        End If
        If world.size.Height > Size.Height Then
            offset.Y += moveY
            If offset.Y < Size.Height - world.size.Height Then offset.Y = Size.Height - world.size.Height
            If offset.Y > 0 Then offset.Y = 0
        End If
        lastMousePoint = e.Location
        setView()
    End Sub

    Private Sub WorldControl_MouseUp(sender As Object, e As MouseEventArgs) Handles Me.MouseUp
        If world.GamePaused OrElse world.MessageVisible Then Return
        If e.Button = MouseButtons.Middle Then dragging = False
    End Sub

    Private Sub setOffset()
        Dim windowSize As Size = Game.window.getClientRect.Size
        If world.size.Width < windowSize.Width Then
            offset.X = (windowSize.Width - world.size.Width) \ 2
        Else
            If offset.X < windowSize.Width - world.size.Width Then
                offset.X = windowSize.Width - world.size.Width
            Else
                offset.X = 0
            End If
        End If
        If world.size.Height < windowSize.Height Then
            offset.Y = (windowSize.Height - world.size.Height) \ 2
        Else
            If offset.Y < windowSize.Height - world.size.Height Then
                offset.Y = windowSize.Height - world.size.Height
            Else
                offset.Y = 0
            End If
        End If
        setView()
    End Sub

    Private Sub setView()
        Dim rect As New Rectangle(-offset.X, -offset.Y, Width, Height)
        map.setView(rect)
    End Sub

    Private Sub WorldControl_Resize(sender As Object, e As EventArgs) Handles Me.Resize
        setOffset()
        sizeControls()
    End Sub

    Private Sub WorldControl_show(sender As Object, e As EventArgs) Handles Me.show
        Dim env As MusicEnvironment = MusicEnvironment.getEnvironment("ingame")
        MusicManager.environment = env
    End Sub

    Private Sub world_currentWaveChanged(sender As Object, e As EventArgs) Handles world.currentWaveChanged
        If infoSidebar IsNot Nothing Then infoSidebar.waveInfo = "Welle " & world.currentWave & " von " & world.waves
    End Sub

    Private Sub world_lifeChanged(sender As Object, e As EventArgs) Handles world.lifeChanged
        If infoSidebar IsNot Nothing Then infoSidebar.life = world.life
    End Sub

    Private Sub world_moneyChanged(sender As Object, e As EventArgs) Handles world.moneyChanged
        If infoSidebar IsNot Nothing Then infoSidebar.money = world.money
    End Sub

    Private Class IngameMenu : Inherits D2DControl
        Private resumeButton As TDButton
        Private leaveButton As TDButton

        Friend Event ResumeGame As EventHandler
        Friend Event LeaveGame As EventHandler

        Sub New(parent As WorldControl)
            D2DParent = parent

            resumeButton = New TDButton
            resumeButton.D2DParent = Me
            resumeButton.Text = Game.guiINI.getValue("game_resume")
            AddHandler resumeButton.Click, Sub() RaiseEvent ResumeGame(Me, EventArgs.Empty)

            leaveButton = New TDButton
            leaveButton.D2DParent = Me
            leaveButton.Text = Game.guiINI.getValue("game_leave")
            AddHandler leaveButton.Click, Sub() RaiseEvent LeaveGame(Me, EventArgs.Empty)

            AddHandler Me.Resize, Sub() ResizeButtons()
            ResizeButtons()
        End Sub

        Private Sub ResizeButtons()
            Dim bWidth = Convert.ToInt32(500 * Game.getWidthFactor)
            Dim bHeight = Convert.ToInt32(50 * Game.getHeightFactor)
            Dim bX = (Width - bWidth) \ 2
            Dim bSize As New Size(bWidth, bHeight)
            resumeButton.Size = bSize
            leaveButton.Size = bSize
            resumeButton.Location = New Point(bX, Convert.ToInt32(300 * Game.getHeightFactor))
            leaveButton.Location = New Point(bX, Convert.ToInt32(370 * Game.getHeightFactor))
        End Sub

        Protected Overrides Sub OnDraw(e As D2DDrawEventArgs)

        End Sub
    End Class

    Private Class WonLostMenu : Inherits D2DControl
        Private label As D2DLabel
        Private button As TDButton

        Friend Event LeaveGame As EventHandler

        Sub New(parent As WorldControl, text As String)
            D2DParent = parent

            label = New D2DLabel(Me)
            label.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)
            label.Text = text

            button = New TDButton
            button.D2DParent = Me
            button.Text = Game.guiINI.getValue("game_leave")
            AddHandler button.Click, Sub() RaiseEvent LeaveGame(Me, EventArgs.Empty)

            AddHandler Me.Resize, Sub() ResizeControls()
            ResizeControls()
        End Sub

        Private Sub ResizeControls()
            Dim bWidth = Convert.ToInt32(500 * Game.getWidthFactor)
            Dim bHeight = Convert.ToInt32(50 * Game.getHeightFactor)
            Dim bX = (Width - bWidth) \ 2
            button.Size = New Size(bWidth, bHeight)
            button.Location = New Point(bX, Convert.ToInt32(330 * Game.getHeightFactor))
            label.Size = New Size(Width, Convert.ToInt32(200 * Game.getHeightFactor))
            label.Font = FontManager.GetFont("headerFormat")
        End Sub

        Protected Overrides Sub OnDraw(e As D2DDrawEventArgs)

        End Sub
    End Class
End Class