﻿Imports System.IO
Imports D2DControlset
Imports SlimDX
Imports SlimDX.Direct2D
Imports SlimDX.DirectWrite
Imports Helpers.Mathematics
Imports System.Reflection

Friend Class WorldEditControl : Inherits Menu
    Private way As List(Of MetaData)
    Private mapWidth, mapHeight As Integer
    Private description As String
    Private name As String
    Private openSettings As Boolean
    Private diag As WorldEditSettingsDialog
    Private offset As Point
    Private dragging As Boolean
    Private lastMousePoint As Point
    Private Shared worldEditINI As INIFile
    Private selectionBar As ToolSelection
    Private currentTool As Integer
    Private waveDiag As WaveEditorDialog
    Private money As Integer
    Private towerDiag As TowerDialog
    Private allowedTowers As UInteger
    Private segmentTypeTrackBar As TDTrackBar
    Private segmentTypeLabel As D2DLabel
    Private selectedSegment As Integer

    Private exitButton As IngameButton
    Private menuButton As IngameButton
    Private saveButton As IngameButton
    Private editButton As IngameButton
    Private wavesButton As IngameButton
    Private towersButton As IngameButton

    Friend Shared ReadOnly Property devMode As Boolean
        Get
            Return Boolean.Parse(worldEditINI.getValue("enableDeveloperMode", Boolean.FalseString))
        End Get
    End Property

    Friend Shared Sub init()
        worldEditINI = New INIFile(Path.Combine(Game.getEnvironmentPath("settings"), "worldEdit.ini"))
    End Sub

    Sub New(name As String)

        'Neue Welten berücksichtigen
        If name = Game.guiINI.getValue("entry_NEW") Then name = getNextWorldName()

        'Standardeinträge festlegen
        Visible = True
        description = worldEditINI.getValue("defaultMapDescription")
        mapWidth = Integer.Parse(worldEditINI.getValue("defaultMapWidth", "0"))
        mapHeight = Integer.Parse(worldEditINI.getValue("defaultMapHeight", "0"))
        Me.name = name
        allowedTowers = &HFFFFFFFFUI

        'Listen initialisieren
        way = New List(Of MetaData)
        Dim info As New List(Of Byte())

        'Map laden
        openSettings = Not load(name, way, info)
        Dim spawnInf = parseAll(info)

        'WellenEditor erstellen
        waveDiag = New WaveEditorDialog(Me, spawnInf)

        'Einstellungsdialog erstellen
        diag = New WorldEditSettingsDialog(name, description, mapWidth, mapHeight, money)
        diag.parentMenu = Me
        AddHandler diag.hide, Sub()
                                  Me.name = diag.name
                                  description = diag.desc
                                  mapWidth = diag.mapWidth
                                  mapHeight = diag.mapHeight
                                  money = diag.money
                                  setOffset()
                              End Sub

        'Turmeditor erstellen
        towerDiag = New TowerDialog(Me, allowedTowers)
        AddHandler towerDiag.hide, Sub() allowedTowers = towerDiag.GetAllowedTowers

        'Toolbox erstellen
        currentTool = -1
        selectionBar = New ToolSelection(Me)
        AddHandler selectionBar.toolChanged, Sub() currentTool = selectionBar.toolIndex

        'Steuerelemente erstellen
        exitButton = New IngameButton
        exitButton.D2DParent = Me
        exitButton.Text = Game.guiINI.getValue("leaveWE")
        AddHandler exitButton.Click, Sub() goUp()

        menuButton = New IngameButton
        menuButton.D2DParent = Me
        menuButton.Text = Game.guiINI.getValue("menuWE")
        AddHandler menuButton.Click, Sub()
                                         saveButton.Visible = Not saveButton.Visible
                                         editButton.Visible = Not editButton.Visible
                                         wavesButton.Visible = Not wavesButton.Visible
                                         towersButton.Visible = Not towersButton.Visible
                                     End Sub

        saveButton = New IngameButton
        saveButton.D2DParent = Me
        saveButton.Text = Game.guiINI.getValue("saveWE")
        saveButton.Visible = False
        AddHandler saveButton.Click, Sub() save()

        editButton = New IngameButton
        editButton.D2DParent = Me
        editButton.Text = Game.guiINI.getValue("propertiesWE")
        editButton.Visible = False
        AddHandler editButton.Click, Sub() Game.changeMenu(diag)

        wavesButton = New IngameButton
        wavesButton.D2DParent = Me
        wavesButton.Text = Game.guiINI.getValue("wavesWE")
        wavesButton.Visible = False
        AddHandler wavesButton.Click, Sub() Game.changeMenu(waveDiag)

        towersButton = New IngameButton
        towersButton.D2DParent = Me
        towersButton.Text = Game.guiINI.getValue("towersWE")
        towersButton.Visible = False
        AddHandler towersButton.Click, Sub() Game.changeMenu(towerDiag)

        segmentTypeLabel = New D2DLabel(Me)
        segmentTypeLabel.TextColor = DirectCast(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)
        segmentTypeLabel.Text = String.Empty

        segmentTypeTrackBar = New TDTrackBar
        segmentTypeTrackBar.D2DParent = Me
        segmentTypeTrackBar.MaxValue = WaySegment.GetAllWaySegments.Length - 1
        segmentTypeTrackBar.Visible = False
        AddHandler segmentTypeTrackBar.Scroll, Sub() way(selectedSegment).extra = Convert.ToByte(segmentTypeTrackBar.Value)
        AddHandler segmentTypeTrackBar.ValueChanged, Sub() segmentTypeLabel.Text = Game.guiINI.getValue("way_type").Replace("%type%", Game.guiINI.getValue("way" & segmentTypeTrackBar.Value & "_Name"))

        resizeControls()
        setOffset()
        AddHandler Me.Resize, Sub()
                                  resizeControls()
                                  setOffset()
                              End Sub

    End Sub

    Private Function getNextWorldName() As String
        Dim dir As New DirectoryInfo(getEnvironmentPath("maps"))
        Dim num = 1
        Dim getName As Func(Of String) = Function() worldEditINI.getValue("defaultMapName").Replace("%num%", num.ToString)
        Dim testFile As New FileInfo(Path.Combine(dir.FullName, getName()))
        Do While testFile.Exists
            num += 1
            testFile = New FileInfo(Path.Combine(dir.FullName, getName()))
        Loop
        Return getName()
    End Function

    Private Sub setOffset()
        Dim windowSize = Game.window.getClientRect.Size
        If mapWidth < windowSize.Width Then
            offset.X = (windowSize.Width - mapWidth) \ 2
        Else
            If offset.X < windowSize.Width - mapWidth Then
                offset.X = windowSize.Width - mapWidth
            Else
                offset.X = 0
            End If
        End If
        If mapHeight < windowSize.Height Then
            offset.Y = (windowSize.Height - mapHeight) \ 2
        Else
            If offset.Y < windowSize.Height - mapHeight Then
                offset.Y = windowSize.Height - mapHeight
            Else
                offset.Y = 0
            End If
        End If
    End Sub

    Private Iterator Function parseAll(info As List(Of Byte())) As IEnumerable(Of SpawnInfo)
        For i = 0 To info.Count - 1
            Dim sI As SpawnInfo = Nothing
            If parseBytes(info.Item(i), sI) Then Yield sI
        Next
    End Function

    Private Function parseBytes(bytes As Byte(), ByRef sI As SpawnInfo) As Boolean
        Dim index = bytes(0)
        Select Case index
            Case 0
                Return False
            Case 1
                mapWidth = BitConverter.ToInt32(bytes, 1)
                mapHeight = BitConverter.ToInt32(bytes, 5)
            Case 2
                Dim chars() = System.Text.Encoding.UTF8.GetChars(bytes, 1, bytes.Length - 1)
                description = New String(chars)
            Case 3
                sI = New SpawnInfo
                sI.time = BitConverter.ToInt32(bytes, 1)
                Dim spawns(bytes.Length - 6) As Byte
                Array.Copy(bytes, 5, spawns, 0, spawns.Length)
                sI.spawns = spawns
                Return True
            Case 4
                If Not (BitConverter.ToBoolean(bytes, 1) OrElse devMode) Then Throw New MapNotCustomException
            Case 5
                money = BitConverter.ToInt32(bytes, 1)
            Case 6
                allowedTowers = BitConverter.ToUInt32(bytes, 1)
        End Select
        Return False
    End Function

    Private Function load(name As String, ByRef way As List(Of MetaData), ByRef info As List(Of Byte())) As Boolean
        Dim di As New DirectoryInfo(Path.Combine(Game.getEnvironmentPath("maps"), name))
        If di.Exists Then
            Dim worldVals = loadWorld(di)
            info.AddRange(worldVals)
            Dim pathVals = loadPath(di)
            way.AddRange(pathVals)
            Return True
        End If
        Return False
    End Function

    Private Iterator Function loadWorld(di As DirectoryInfo) As IEnumerable(Of Byte())
        Dim fi As New FileInfo(Path.Combine(di.FullName, "map.td1"))
        Dim bytes() As Byte
        Using fs = fi.OpenRead
            bytes = New Byte(Convert.ToInt32(fs.Length) - 1) {}
            Using br As New BinaryReader(fs)
                For i = 0 To bytes.Length - 1
                    bytes(i) = br.ReadByte
                Next
            End Using
        End Using
        Dim seperator As Byte = &H7F
        Dim index = 0
        Do While index < bytes.Length
            Dim bList As New List(Of Byte)
            Do While index < bytes.Length AndAlso Not bytes(index) = seperator
                bList.Add(bytes(index))
                index += 1
            Loop
            Yield bList.ToArray
            index += 1
        Loop
    End Function

    Private Iterator Function loadPath(di As DirectoryInfo) As IEnumerable(Of MetaData)
        Dim fi As New FileInfo(Path.Combine(di.FullName, "map.td2"))
        Using fs = fi.OpenRead
            Using br As New BinaryReader(fs)
                For i = 1 To Convert.ToInt32(fs.Length / 9)
                    Dim data As New MetaData
                    data.pX = br.ReadInt32
                    data.pY = br.ReadInt32
                    data.extra = br.ReadByte()
                    Yield data
                Next
            End Using
        End Using
    End Function

    Private Class MetaData
        Friend pX As Integer
        Friend pY As Integer
        Friend extra As Byte
    End Class

    Private Sub save()
        Dim di As New DirectoryInfo(Path.Combine(Game.getEnvironmentPath("maps"), name))
        If Not di.Exists Then di.Create()
        saveWorld(di)
        savePath(di)
        Game.reloadWorlds()
        Game.refreshMapList()
        Game.refreshEditorList()
    End Sub

    Private Sub saveWorld(di As DirectoryInfo)
        Dim fi As New FileInfo(Path.Combine(di.FullName, "map.td1"))
        Dim seperator As Byte = &H7F
        Using fs = fi.Open(FileMode.Create, FileAccess.Write)
            Using bw As New BinaryWriter(fs)
                bw.Write(CByte(1))
                bw.Write(mapWidth)
                bw.Write(mapHeight)
                bw.Write(seperator)
                bw.Write(CByte(2))
                bw.Write(System.Text.Encoding.UTF8.GetBytes(description))
                For Each sI In waveDiag.GetSpawns
                    bw.Write(seperator)
                    bw.Write(CByte(3))
                    bw.Write(sI.time)
                    bw.Write(sI.spawns)
                Next
                bw.Write(seperator)
                bw.Write(CByte(4))
                bw.Write(Not devMode)
                bw.Write(seperator)
                bw.Write(CByte(5))
                bw.Write(money)
                bw.Write(seperator)
                bw.Write(CByte(6))
                bw.Write(allowedTowers)
            End Using
        End Using
    End Sub

    Private Sub savePath(di As DirectoryInfo)
        Dim fi As New FileInfo(Path.Combine(di.FullName, "map.td2"))
        Using fs = fi.Open(FileMode.Create, FileAccess.Write)
            Using bw As New BinaryWriter(fs)
                For i = 0 To way.Count - 1
                    Dim data = way(i)
                    bw.Write(data.pX)
                    bw.Write(data.pY)
                    bw.Write(data.extra)
                Next
            End Using
        End Using
    End Sub

    Protected Overrides Sub OnDraw(e As D2DDrawEventArgs)
        Dim realOffset As New Point(offset.X + AbsoluteLocation.X, offset.Y + AbsoluteLocation.Y)
        Dim mapRect As New Rectangle(realOffset.X - 1, realOffset.Y - 1, mapWidth + 2, mapHeight + 2)
        Dim drawToolAllowed = mapRect.Contains(PointToClient(Cursor.Position))
        e.Target.PushAxisAlignedClip(mapRect, AntialiasMode.PerPrimitive)
        drawWay(e, realOffset)
        drawMapBorder(e, realOffset)
        If drawToolAllowed Then drawTool(e, realOffset)
        e.Target.PopAxisAlignedClip()
    End Sub

    Private Sub drawTool(e As D2DDrawEventArgs, offset As Point)
        If currentTool < 0 Then Exit Sub
        Dim mousePos = PointToClient(Cursor.Position)
        mousePos.X += AbsoluteLocation.X
        mousePos.Y += AbsoluteLocation.Y
        Select Case currentTool
            Case 0
                Dim b As New SolidColorBrush(e.Target, New Color4(Color.DimGray))
                b.Opacity = 0.5F
                If way.Count > 0 Then
                    If (Control.ModifierKeys And Keys.Shift) = Keys.Shift Then
                        If Math.Abs(mousePos.X - (way.Last.pX + offset.X)) > Math.Abs(mousePos.Y - (way.Last.pY + offset.Y)) Then
                            mousePos.Y = way.Last.pY + offset.Y
                        Else
                            mousePos.X = way.Last.pX + offset.X
                        End If
                    End If
                    Dim pg As New PathGeometry(e.Factory)
                    Dim sink = pg.Open
                    sink.BeginFigure(New Point(way.Last.pX + offset.X, way.Last.pY + offset.Y), FigureBegin.Hollow)
                    sink.AddLine(mousePos)
                    sink.EndFigure(FigureEnd.Open)
                    sink.Close()
                    Dim sProps As New StrokeStyleProperties
                    sProps.DashCap = CapStyle.Round
                    sProps.DashStyle = DashStyle.Solid
                    sProps.EndCap = CapStyle.Round
                    sProps.StartCap = CapStyle.Round
                    sProps.LineJoin = LineJoin.Round
                    Dim style As New StrokeStyle(e.Factory, sProps)
                    e.Target.DrawGeometry(pg, b, 24.0F, style)
                    style.Dispose()
                    sink.Dispose()
                    pg.Dispose()
                Else
                    Dim endCircle As New Ellipse
                    endCircle.Center = mousePos
                    endCircle.RadiusX = 12.0F
                    endCircle.RadiusY = 12.0F
                    e.Target.FillEllipse(b, endCircle)
                End If
                b.Dispose()
            Case 1
                Dim i = getWayIndex(New Point(mousePos.X - offset.X, mousePos.Y - offset.Y), True)
                If i >= 0 Then
                    Dim b1 As New SolidColorBrush(e.Target, New Color4(Color.DimGray))
                    b1.Opacity = 0.5F
                    Dim b2 As New SolidColorBrush(e.Target, New Color4(Color.Orange))
                    b2.Opacity = 0.5F
                    Dim pg As New PathGeometry(e.Factory)
                    Dim sink = pg.Open
                    Dim startIndex = Math.Max(i - 1, 0)
                    Dim startPoint = New Point(way(startIndex).pX + offset.X, way(startIndex).pY + offset.Y)
                    sink.BeginFigure(startPoint, FigureBegin.Hollow)
                    If startIndex < way.Count - 1 Then sink.AddLine(New Point(way(startIndex + 1).pX + offset.X, way(startIndex + 1).pY + offset.Y))
                    If i < way.Count - 1 AndAlso i > 0 Then sink.AddLine(New Point(way(startIndex + 2).pX + offset.X, way(startIndex + 2).pY + offset.Y))
                    sink.EndFigure(FigureEnd.Open)
                    sink.Close()
                    Dim sProps As New StrokeStyleProperties
                    sProps.DashCap = CapStyle.Round
                    sProps.DashStyle = DashStyle.Solid
                    sProps.EndCap = CapStyle.Round
                    sProps.StartCap = CapStyle.Round
                    sProps.LineJoin = LineJoin.Round
                    Dim style As New StrokeStyle(e.Factory, sProps)
                    If i < way.Count - 1 AndAlso i > 0 Then
                        Dim pg2 As New PathGeometry(e.Factory)
                        Dim sink2 = pg2.Open
                        sink2.BeginFigure(New Point(way(i - 1).pX + offset.X, way(i - 1).pY + offset.Y), FigureBegin.Hollow)
                        sink2.AddLine(New Point(way(i + 1).pX + offset.X, way(i + 1).pY + offset.Y))
                        sink2.EndFigure(FigureEnd.Open)
                        sink2.Close()
                        e.Target.DrawGeometry(pg2, b1, 24.0F, style)
                        sink2.Dispose()
                        pg2.Dispose()
                    End If
                    e.Target.DrawGeometry(pg, b2, 24.0F, style)
                    style.Dispose()
                    sink.Dispose()
                    pg.Dispose()
                    b2.Dispose()
                    b1.Dispose()
                End If
            Case 2
                Dim i = getWayIndex(New Point(mousePos.X - offset.X, mousePos.Y - offset.Y), False)
                If i >= 0 Then
                    Dim b As New SolidColorBrush(e.Target, New Color4(Color.Orange))
                    b.Opacity = 0.5F
                    Dim pg As New PathGeometry(e.Factory)
                    Dim sink = pg.Open
                    sink.BeginFigure(New Point(way(i - 1).pX + offset.X, way(i - 1).pY + offset.Y), FigureBegin.Hollow)
                    sink.AddLine(New Point(way(i).pX + offset.X, way(i).pY + offset.Y))
                    sink.EndFigure(FigureEnd.Open)
                    sink.Close()
                    Dim sProps As New StrokeStyleProperties
                    sProps.DashCap = CapStyle.Round
                    sProps.DashStyle = DashStyle.Solid
                    sProps.EndCap = CapStyle.Round
                    sProps.StartCap = CapStyle.Round
                    sProps.LineJoin = LineJoin.Round
                    Dim style As New StrokeStyle(e.Factory, sProps)
                    e.Target.DrawGeometry(pg, b, 24.0F, style)
                    style.Dispose()
                    sink.Dispose()
                    pg.Dispose()
                    b.Dispose()
                End If
        End Select
    End Sub

    Private Sub drawWay(e As D2DDrawEventArgs, offset As Point)
        Dim b As New SolidColorBrush(e.Target, New Color4(Color.DimGray))
        Dim getPoint As Func(Of MetaData, PointF) = Function(item) New Point(offset.X + item.pX, offset.Y + item.pY)
        If way.Count > 1 Then
            Dim pg As New PathGeometry(e.Factory)
            Dim sink = pg.Open
            sink.BeginFigure(getPoint(way(0)), FigureBegin.Hollow)
            sink.AddLines((From p In way Skip 1 Select getPoint(p)).ToArray)
            sink.EndFigure(FigureEnd.Open)
            sink.Close()
            Dim sProps As New StrokeStyleProperties
            sProps.DashCap = CapStyle.Round
            sProps.DashStyle = DashStyle.Solid
            sProps.EndCap = CapStyle.Round
            sProps.StartCap = CapStyle.Round
            sProps.LineJoin = LineJoin.Round
            Dim style As New StrokeStyle(e.Factory, sProps)
            e.Target.DrawGeometry(pg, b, 24.0F, style)
            style.Dispose()
            sink.Dispose()
            pg.Dispose()
        End If
        If way.Count > 0 Then
            Dim endCircle As New Ellipse
            endCircle.Center = getPoint(way.Last)
            endCircle.RadiusX = 18
            endCircle.RadiusY = 18
            e.Target.FillEllipse(b, endCircle)
        End If
        b.Dispose()
    End Sub

    Private Sub drawMapBorder(e As D2DDrawEventArgs, offset As Point)
        Dim sProps As New StrokeStyleProperties
        sProps.DashStyle = DashStyle.Custom
        Dim style As New StrokeStyle(e.Factory, sProps, {5, 8})
        e.Target.DrawRectangle(ColorManager.GetBrush("defaultWhite"), New RectangleF(offset.X - 0.5F, offset.Y - 0.5F, mapWidth + 1, mapHeight + 1), 1.0F, style)
        style.Dispose()
    End Sub

    Private Sub WorldEditControl_MouseClick(sender As Object, e As MouseEventArgs) Handles Me.MouseClick
        Dim mapRect As New Rectangle(AbsoluteLocation.X + offset.X - 1, AbsoluteLocation.Y + offset.Y - 1, mapWidth + 2, mapHeight + 2)
        Dim mousePos = PointToClient(Cursor.Position)
        Dim checkRect = segmentTypeTrackBar.ClientRectangle
        Dim rectBorder = Convert.ToInt32(40 * Game.getSizeFactor)
        checkRect.X -= rectBorder
        checkRect.Y -= rectBorder
        checkRect.Width += 2 * rectBorder
        checkRect.Height += 2 * rectBorder
        If segmentTypeTrackBar.Visible AndAlso Not checkRect.Contains(e.Location) Then
            segmentTypeTrackBar.Visible = False
            segmentTypeLabel.Visible = False
        End If
        If Not (mapRect.Contains(mousePos) AndAlso e.Button = MouseButtons.Left) Then Return
        Dim posOnMap = New Point(mousePos.X - offset.X, mousePos.Y - offset.Y)
        Select Case currentTool
            Case 0
                If way.Count > 0 AndAlso (Control.ModifierKeys And Keys.Shift) = Keys.Shift Then
                    If Math.Abs(posOnMap.X - way.Last.pX) > Math.Abs(posOnMap.Y - way.Last.pY) Then
                        posOnMap.Y = way.Last.pY
                    Else
                        posOnMap.X = way.Last.pX
                    End If
                End If
                If way.Count > 0 Then way.Last.extra = 0
                Dim data As New MetaData
                data.extra = 1
                data.pX = posOnMap.X
                data.pY = posOnMap.Y
                way.Add(data)
            Case 1
                Dim i = getWayIndex(posOnMap, True)
                If i >= 0 Then way.RemoveAt(i)
            Case 2
                Dim i = getWayIndex(posOnMap, False)
                If i < way.Count - 1 AndAlso i >= 0 Then
                    segmentTypeTrackBar.Visible = True
                    segmentTypeLabel.Visible = True
                    selectedSegment = i
                    segmentTypeTrackBar.Value = way(i).extra
                End If
        End Select
    End Sub

    Private Function getWayIndex(p As Point, first As Boolean) As Integer
        If way.Count <= 0 Then Return -1
        Dim v As New Helpers.Mathematics.Vector2(p)
        If first AndAlso MathHelper.pointInCircle(New Helpers.Mathematics.Vector2(way.First.pX, way.First.pY), 12, v) Then Return 0
        For i = 1 To way.Count - 1
            Dim points() = getPoints(i, 1)
            For j = 0 To points.Length - 1
                Dim pVec As New Helpers.Mathematics.Vector2(points(j))
                If MathHelper.pointInCircle(pVec, 12, v) Then Return i
            Next
        Next
        Return -1
    End Function

    Private Function getPoints(i As Integer, distance As Double) As PointF()
        If i < 1 Then Return Nothing
        Dim seg1 = way(i - 1)
        Dim seg2 = way(i)
        Dim steps = Convert.ToInt32(MathHelper.getDistance(New Helpers.Mathematics.Vector2(seg1.pX, seg1.pY), New Helpers.Mathematics.Vector2(seg2.pX, seg2.pY)) / distance)
        If steps > 0 Then
            Dim p(steps) As PointF
            p(0) = New PointF(seg1.pX, seg1.pY)
            Dim sX = Convert.ToSingle((seg2.pX - seg1.pX) / steps)
            Dim sY = Convert.ToSingle((seg2.pY - seg1.pY) / steps)
            For i = 1 To steps
                p(i) = New PointF(seg1.pX + i * sX, seg1.pY + i * sY)
            Next
            Return p
        End If
        Return Nothing
    End Function

    Private Sub WorldEditControl_MouseDown(sender As Object, e As MouseEventArgs) Handles Me.MouseDown
        If (DateTime.Now - Game.lastMenuChange).Ticks < 100000 OrElse e.Button <> MouseButtons.Left Then Return
        If currentTool < 0 Then
            dragging = True
            lastMousePoint = e.Location
        End If
    End Sub

    Private Sub WorldEditControl_MouseMove(sender As Object, e As MouseEventArgs) Handles Me.MouseMove
        If Not dragging Then Exit Sub
        Dim moveX As Integer = e.X - lastMousePoint.X
        Dim moveY As Integer = e.Y - lastMousePoint.Y
        If mapWidth > Size.Width Then
            offset.X += moveX
            If offset.X < Size.Width - mapWidth Then offset.X = Size.Width - mapWidth
            If offset.X > 0 Then offset.X = 0
        End If
        If mapHeight > Size.Height Then
            offset.Y += moveY
            If offset.Y < Size.Height - mapHeight Then offset.Y = Size.Height - mapHeight
            If offset.Y > 0 Then offset.Y = 0
        End If
        lastMousePoint = e.Location
    End Sub

    Private Sub WorldEditControl_MouseUp(sender As Object, e As MouseEventArgs) Handles Me.MouseUp
        dragging = False
    End Sub

    Private Sub WorldEditControl_show(sender As Object, e As EventArgs) Handles Me.show
        If openSettings Then
            openSettings = False
            Game.changeMenu(diag)
        End If
    End Sub

    Private Sub resizeControls()
        Dim bWidth = Convert.ToInt32(170 * Game.getWidthFactor)
        Dim bHeight = Convert.ToInt32(34 * Game.getHeightFactor)
        Dim bBorder = Convert.ToInt32(10 * Game.getSizeFactor)
        Dim bSize As New Size(bWidth, bHeight)
        exitButton.Size = bSize
        menuButton.Size = bSize
        saveButton.Size = bSize
        editButton.Size = bSize
        wavesButton.Size = bSize
        towersButton.Size = bSize
        Dim bX = (Width - bWidth) \ 2
        exitButton.Location = New Point(bBorder, bBorder)
        menuButton.Location = New Point(bX, bBorder)
        saveButton.Location = New Point(bX, 3 * bBorder + bHeight)
        editButton.Location = New Point(bX, 4 * bBorder + 2 * bHeight)
        wavesButton.Location = New Point(bX, 5 * bBorder + 3 * bHeight)
        towersButton.Location = New Point(bX, 6 * bBorder + 4 * bHeight)
        selectionBar.Location = New Point(Width - bBorder - Convert.ToInt32(Game.getSizeFactor * 60), bBorder)
        segmentTypeLabel.Font = FontManager.GetFont("defaultLabelFormat")
        segmentTypeTrackBar.Size = New Size(Convert.ToInt32(600 * Game.getWidthFactor), Convert.ToInt32(25 * Game.getHeightFactor))
        segmentTypeTrackBar.Location = New Point((Width - segmentTypeTrackBar.Width) \ 2, Convert.ToInt32(600 * Game.getHeightFactor))
        segmentTypeLabel.Size = New Size(segmentTypeTrackBar.Width, Convert.ToInt32(50 * Game.getHeightFactor))
        segmentTypeLabel.Location = New Point(segmentTypeTrackBar.X, segmentTypeTrackBar.Y - segmentTypeLabel.Height)
    End Sub

    Private Class ToolSelection : Inherits GameControl
        Private index As Integer

        Friend Event toolChanged(sender As Object, e As EventArgs)

        Friend ReadOnly Property toolIndex As Integer
            Get
                Return index
            End Get
        End Property

        Private ReadOnly Property itemSize As Integer
            Get
                Return Convert.ToInt32(30 * Game.getSizeFactor)
            End Get
        End Property

        Protected Overrides Sub drawContent(e As D2DDrawEventArgs, inRect As RectangleF)
            Dim absLoc = AbsoluteLocation
            Dim iSize = itemSize
            Dim x = absLoc.X + iSize \ 2
            For i = 0 To 2
                Dim rect As New Rectangle(x, absLoc.Y + ((i + 1) * iSize \ 2) + i * iSize, iSize, iSize)
                If i = index Then
                    Dim b As New SolidColorBrush(e.Target, CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush).Color)
                    b.Opacity = 0.3F
                    Dim rBorder = iSize \ 4
                    e.Target.FillRectangle(b, New Rectangle(rect.X - rBorder, rect.Y - rBorder, rect.Width + 2 * rBorder, rect.Height + 2 * rBorder))
                    b.Dispose()
                End If
                e.Target.DrawBitmap(Game.toolImages(i), rect)
            Next
        End Sub

        Sub New(parent As WorldEditControl)
            MyBase.New(parent)
            index = -1
            resizeSelf()
            AddHandler m_Parent.Resize, Sub() resizeSelf()
        End Sub

        Private Sub resizeSelf()
            Width = 2 * itemSize
            Height = 5 * itemSize
        End Sub

        Private Sub ToolSelection_MouseClick(sender As Object, e As MouseEventArgs) Handles Me.MouseClick
            Dim iSize = itemSize
            For i = 0 To 2
                Dim rect As New Rectangle(iSize \ 2, ((i + 1) * iSize \ 2) + i * iSize, iSize, iSize)
                If rect.Contains(e.Location) Then
                    index = i
                    RaiseEvent toolChanged(Me, EventArgs.Empty)
                    Exit For
                End If
            Next
        End Sub

        Private Sub parent_MouseClick(sender As Object, e As MouseEventArgs) Handles m_Parent.MouseClick
            If e.Button = MouseButtons.Right Then
                index = -1
                RaiseEvent toolChanged(Me, EventArgs.Empty)
            End If
        End Sub
    End Class

    Private Class WorldEditSettingsDialog : Inherits Menu
        Private nameTextBox As TDTextBox
        Private moneyTextBox As TDTextBox
        Private descTextBox As TDTextBox
        Private widthTextBox As TDTextBox
        Private heightTextBox As TDTextBox
        Private button As TDButton
        Private label1 As D2DLabel
        Private label2 As D2DLabel
        Private label3 As D2DLabel
        Private label4 As D2DLabel
        Private label5 As D2DLabel

        Friend ReadOnly Property name As String
            Get
                Return nameTextBox.Text
            End Get
        End Property

        Friend ReadOnly Property desc As String
            Get
                Return descTextBox.Text
            End Get
        End Property

        Friend ReadOnly Property mapWidth As Integer
            Get
                Return Integer.Parse(widthTextBox.Text)
            End Get
        End Property

        Friend ReadOnly Property mapHeight As Integer
            Get
                Return Integer.Parse(heightTextBox.Text)
            End Get
        End Property

        Friend ReadOnly Property money As Integer
            Get
                Return Integer.Parse(moneyTextBox.Text)
            End Get
        End Property

        Sub New(name As String, desc As String, mapWidth As Integer, mapHeight As Integer, money As Integer)
            nameTextBox = New TDTextBox(Me)
            nameTextBox.MultiLine = False
            nameTextBox.Text = name

            moneyTextBox = New TDTextBox(Me)
            moneyTextBox.MultiLine = False
            moneyTextBox.Text = money.ToString
            moneyTextBox.IsNumeric = True

            descTextBox = New TDTextBox(Me)
            descTextBox.MultiLine = True
            descTextBox.Text = desc

            widthTextBox = New TDTextBox(Me)
            widthTextBox.IsNumeric = True
            widthTextBox.Text = mapWidth.ToString
            widthTextBox.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)

            heightTextBox = New TDTextBox(Me)
            heightTextBox.IsNumeric = True
            heightTextBox.Text = mapHeight.ToString
            heightTextBox.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)

            button = New TDButton
            button.D2DParent = Me
            button.Text = Game.guiINI.getValue("doneWE")
            AddHandler button.Click, Sub() goUp()

            label1 = New D2DLabel(Me)
            label1.Text = Game.guiINI.getValue("nameWE")
            label1.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)

            label2 = New D2DLabel(Me)
            label2.Text = Game.guiINI.getValue("descriptionWE")
            label2.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)

            label3 = New D2DLabel(Me)
            label3.Text = Game.guiINI.getValue("widthWE")
            label3.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)

            label4 = New D2DLabel(Me)
            label4.Text = Game.guiINI.getValue("heightWE")
            label4.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)

            label5 = New D2DLabel(Me)
            label5.Text = Game.guiINI.getValue("moneyWE")
            label5.TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)

            resizeControls()
            AddHandler Me.Resize, Sub() resizeControls()
        End Sub

        Private Sub resizeControls()
            label1.Font = FontManager.GetFont("defaultLabelFormat")
            label2.Font = FontManager.GetFont("defaultLabelFormat")
            label3.Font = FontManager.GetFont("defaultLabelFormat")
            label4.Font = FontManager.GetFont("defaultLabelFormat")
            label5.Font = FontManager.GetFont("defaultLabelFormat")

            Dim cWidth = Convert.ToInt32(600 * Game.getWidthFactor)
            Dim cX = (Width - cWidth) \ 2
            Dim textBoxBorder = Convert.ToInt32(60 * Game.getHeightFactor)
            Dim textBoxHeight = Convert.ToInt32(40 * Game.getHeightFactor)

            button.Location = New Point(cX, Convert.ToInt32(600 * Game.getHeightFactor))
            button.Size = New Size(cWidth, Convert.ToInt32(50 * Game.getHeightFactor))

            nameTextBox.Location = New Point(cX, Convert.ToInt32(100 * Game.getHeightFactor))
            nameTextBox.Size = New Size(cWidth, textBoxHeight)
            label1.Location = New Point(cX, nameTextBox.Y - textBoxHeight)
            label1.Size = New Size(cWidth, nameTextBox.Y - label1.Y)

            moneyTextBox.Location = New Point(cX, textBoxBorder + nameTextBox.Y + nameTextBox.Height)
            moneyTextBox.Size = New Size(cWidth, textBoxHeight)
            label5.Location = New Point(cX, moneyTextBox.Y - textBoxHeight)
            label5.Size = New Size(cWidth, moneyTextBox.Y - label5.Y)

            descTextBox.Location = New Point(cX, textBoxBorder + moneyTextBox.Y + moneyTextBox.Height)
            descTextBox.Size = New Size(cWidth, Convert.ToInt32(180 * Game.getHeightFactor))
            label2.Location = New Point(cX, descTextBox.Y - textBoxHeight)
            label2.Size = New Size(cWidth, descTextBox.Y - label2.Y)

            Dim udWidth = Convert.ToInt32(cWidth * 0.4) \ 2
            Dim udY = Convert.ToInt32(40 * Game.getHeightFactor) + descTextBox.Y + descTextBox.Height
            Dim udHeight = Convert.ToInt32(40 * Game.getHeightFactor)

            widthTextBox.Location = New Point(cX + udWidth, udY)
            heightTextBox.Location = New Point((cX + cWidth) - udWidth, udY)

            Dim udSize As New Size(udWidth, udHeight)
            widthTextBox.Size = udSize
            heightTextBox.Size = udSize
            label3.Location = New Point(cX, udY)
            label3.Size = udSize
            label4.Location = New Point((cX + cWidth) - 2 * udWidth, udY)
            label4.Size = udSize
        End Sub
    End Class

    Private Class WaveEditorDialog : Inherits Menu
        Private spawnInf As List(Of AdvancedSpawnInfo)
        Private listBox As TDListBox
        Private editDiag As EditDialog
        Private upButton As TDButton
        Private downButton As TDButton
        Private backButton As TDButton
        Private newButton As TDButton
        Private editButton As TDButton
        Private removeButton As TDButton

        Friend Function GetSpawns() As SpawnInfo()
            Return (From spawn In spawnInf Where spawn.Items.Count > 0 Select spawn.ToSpawnInfo).ToArray
        End Function

        Sub New(parent As WorldEditControl, spawns As IEnumerable(Of SpawnInfo))
            parentMenu = parent
            spawnInf = New List(Of AdvancedSpawnInfo)(From sI In spawns Select New AdvancedSpawnInfo(sI))
            listBox = New TDListBox
            listBox.D2DParent = Me
            editDiag = New EditDialog(Me)
            AddHandler editDiag.hide, Sub() RefreshList()

            upButton = New TDButton
            upButton.D2DParent = Me
            upButton.Text = Game.guiINI.getValue("wave_up")
            AddHandler upButton.Click, Sub()
                                           If listBox.SelectedIndex > 0 Then
                                               Dim index = listBox.SelectedIndex
                                               MoveItem(index, index - 1)
                                               listBox.SelectedIndex = index - 1
                                           End If
                                       End Sub

            downButton = New TDButton
            downButton.D2DParent = Me
            downButton.Text = Game.guiINI.getValue("wave_down")
            AddHandler downButton.Click, Sub()
                                             If listBox.SelectedIndex < listBox.Items.Count - 1 Then
                                                 Dim index = listBox.SelectedIndex
                                                 MoveItem(index, index + 1)
                                                 listBox.SelectedIndex = index + 1
                                             End If
                                         End Sub

            backButton = New TDButton
            backButton.D2DParent = Me
            backButton.Text = Game.guiINI.getValue("back")
            AddHandler backButton.Click, Sub() goUp()

            newButton = New TDButton
            newButton.D2DParent = Me
            newButton.Text = Game.guiINI.getValue("wave_new")
            AddHandler newButton.Click, Sub() AddItem()

            editButton = New TDButton
            editButton.D2DParent = Me
            editButton.Text = Game.guiINI.getValue("wave_edit")
            AddHandler editButton.Click, Sub() EditItem(listBox.SelectedIndex)

            removeButton = New TDButton
            removeButton.D2DParent = Me
            removeButton.Text = Game.guiINI.getValue("wave_remove")
            AddHandler removeButton.Click, Sub() RemoveItem(listBox.SelectedIndex)

            AddHandler Me.Resize, Sub() ResizeControls()
            ResizeControls()
            RefreshList()
        End Sub

        Private Sub ResizeControls()
            Dim cWidth = Convert.ToInt32(Width * 0.4)
            Dim cBorderX = Convert.ToInt32(Width * 0.07)
            Dim cBorderY = Convert.ToInt32(Height * 0.1)
            listBox.Size = New Size(cWidth, Convert.ToInt32(Height * 0.8))
            listBox.Location = New Point(cBorderX, cBorderY)
            listBox.ItemHeight = Convert.ToInt32(25 * Game.getSizeFactor)
            Dim buttonSize As New Size(cWidth, Convert.ToInt32(50 * Game.getSizeFactor))
            Dim buttonBorder = Convert.ToInt32(15 * Game.getSizeFactor)
            upButton.Size = buttonSize
            downButton.Size = buttonSize
            backButton.Size = buttonSize
            newButton.Size = buttonSize
            editButton.Size = buttonSize
            removeButton.Size = buttonSize
            Dim buttonX = Width - cBorderX - cWidth
            upButton.Location = New Point(buttonX, cBorderY)
            downButton.Location = New Point(buttonX, cBorderY + buttonSize.Height + buttonBorder)
            backButton.Location = New Point(buttonX, Height - cBorderY - buttonSize.Height)
            newButton.Location = New Point(buttonX, Convert.ToInt32(300 * Game.getHeightFactor))
            editButton.Location = New Point(buttonX, Convert.ToInt32(365 * Game.getHeightFactor))
            removeButton.Location = New Point(buttonX, Convert.ToInt32(430 * Game.getHeightFactor))
        End Sub

        Private Sub RefreshList()
            listBox.Items.Clear()
            spawnInf.Sort(Function(obj1, obj2) obj1.Time.CompareTo(obj2.Time))
            For Each spawn In spawnInf
                Dim item As New WaveListBoxItem
                item.Info = spawn
                listBox.Items.Add(item)
            Next
        End Sub

        Private Sub AddItem()
            EditItem(spawnInf.Count)
        End Sub

        Private Sub EditItem(index As Integer)
            If index < 0 Then Return
            editDiag.Index = index
            Game.changeMenu(editDiag)
        End Sub

        Private Sub RemoveItem(index As Integer)
            If index < 0 Then Return
            spawnInf.RemoveAt(index)
            RefreshList()
        End Sub

        Private Sub MoveItem(index As Integer, destIndex As Integer)
            If index < 0 Then Return
            Dim item = spawnInf(index)
            spawnInf.RemoveAt(index)
            'If destIndex > index Then destIndex -= 1
            spawnInf.Insert(destIndex, item)
            RefreshList()
        End Sub

        Private Class AdvancedSpawnInfo
            Private m_Items As List(Of SpawnInfoItem)

            Friend Property Time As Integer

            Friend ReadOnly Property Items As List(Of SpawnInfoItem)
                Get
                    Return m_Items
                End Get
            End Property

            Sub New(info As SpawnInfo)
                Me.New()
                Time = info.time
                For i = 0 To info.spawns.Length - 1 Step 5
                    Dim item As New SpawnInfoItem
                    item.Type = BitConverter.ToInt32(info.spawns, i)
                    item.Count = info.spawns(i + 4)
                    m_Items.Add(item)
                Next
            End Sub

            Sub New()
                m_Items = New List(Of SpawnInfoItem)
            End Sub

            Friend Function ToSpawnInfo() As SpawnInfo
                Dim sI As New SpawnInfo
                sI.time = Time
                sI.spawns = New Byte(m_Items.Count * 5 - 1) {}
                For i = 0 To m_Items.Count - 1
                    Array.ConstrainedCopy(BitConverter.GetBytes(m_Items(i).Type), 0, sI.spawns, 5 * i, 4)
                    sI.spawns(5 * i + 4) = m_Items(i).Count
                Next
                Return sI
            End Function

            Friend Class SpawnInfoItem
                Friend Property Type As Integer
                Friend Property Count As Byte
            End Class
        End Class

        Private Class WaveListBoxItem : Inherits TDListBox.ListBoxItem

            Friend Property Info As AdvancedSpawnInfo

            Protected Overrides Sub DrawItem(rect As Rectangle, e As D2DDrawEventArgs)
                If IsSelected Then e.Target.DrawRectangle(ColorManager.GetBrush("defaultWhite"), New RectangleF(rect.X + 0.5F, rect.Y + 0.5F, rect.Width - 1, rect.Height - 1))
                If Info Is Nothing Then Return
                Dim timeRect As New Rectangle(rect.X, rect.Y, Convert.ToInt32(120 * Game.getWidthFactor), rect.Height)
                Dim formatName = "ListBoxTailingItemFormat"
                If Not FontManager.FontRegistered(formatName) Then FontManager.RegisterFont(formatName, New FontTemplate("Polentical Neon", 16, TextAlignment:=TextAlignment.Trailing, ParagraphAlignment:=ParagraphAlignment.Center))
                e.Target.DrawText((Info.Time / 60).ToString("0.0") & "sec:", FontManager.GetFont(formatName), timeRect, ColorManager.GetBrush("defaultWhite"))
                Dim counterFormatName = "ListBoxItemCounterFormat"
                If Not FontManager.FontRegistered(counterFormatName) Then FontManager.RegisterFont(counterFormatName, New FontTemplate("Segoe UI", 14, FontStyle.Normal, FontWeight.SemiBold, TextAlignment.Trailing, ParagraphAlignment.Far))
                Dim itemBorder = Convert.ToInt32(10 * Game.getWidthFactor)
                Dim rectSize = Convert.ToInt32(rect.Height * 0.7)
                For i = 0 To Info.Items.Count - 1
                    Dim item = Info.Items(i)
                    Dim creepRect As New Rectangle(rect.X + timeRect.Width + (i + 1) * itemBorder + i * rect.Height + (rect.Height - rectSize) \ 2, rect.Y + (rect.Height - rectSize) \ 2, rectSize, rectSize)
                    CreepRenderer.DrawCreep(item.Type, e, creepRect)
                    Dim counterRect As New Rectangle(rect.X + timeRect.Width + (i + 1) * itemBorder + i * rect.Height, rect.Y, rect.Height, rect.Height)
                    e.Target.DrawText(item.Count.ToString, FontManager.GetFont(counterFormatName), counterRect, ColorManager.GetBrush("defaultWhite"))
                Next
            End Sub
        End Class

        Private Class EditDialog : Inherits Menu
            Private waveDiag As WaveEditorDialog
            Private spawn As AdvancedSpawnInfo
            Private listBox As TDListBox
            Private trackBar1 As TDTrackBar
            Private trackBar2 As TDTrackBar
            Private upButton As TDButton
            Private downButton As TDButton
            Private newButton As TDButton
            Private removeButton As TDButton
            Private backButton As TDButton
            Private label1 As D2DLabel
            Private label2 As D2DLabel
            Private label3 As D2DLabel
            Private textBox As TDTextBox

            Private creeps As CreepDescriptor()

            Friend Property Index As Integer

            Sub New(parent As WaveEditorDialog)
                parentMenu = parent
                waveDiag = parent
                AddHandler Me.show, AddressOf EditDialog_Show
                AddHandler Me.hide, Sub() spawn.Time = Convert.ToInt32(Double.Parse(textBox.Text) * 60)
                creeps = Creep.GetAllCreeps

                label1 = New D2DLabel(Me) With {.Text = String.Empty, .TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush), .Visible = False}
                label2 = New D2DLabel(Me) With {.Text = String.Empty, .TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush), .Visible = False}
                label3 = New D2DLabel(Me) With {.Text = Game.guiINI.getValue("wave_time"), .TextColor = CType(ColorManager.GetBrush("defaultWhite"), SolidColorBrush), .Visible = True}

                textBox = New TDTextBox(Me)
                textBox.IsNumeric = True

                listBox = New TDListBox
                listBox.D2DParent = Me
                AddHandler listBox.SelectedIndexChanged, AddressOf OnSelectedIndexChanged

                trackBar1 = New TDTrackBar
                trackBar1.D2DParent = Me
                trackBar1.MinValue = 0
                trackBar1.MaxValue = creeps.Length - 1
                trackBar1.Visible = False
                AddHandler trackBar1.Scroll, AddressOf OnTrackBar1Scroll
                AddHandler trackBar1.ValueChanged, Sub() label1.Text = Game.guiINI.getValue("creep_type").Replace("%type%", Game.guiINI.getValue("creep" & trackBar1.Value))

                trackBar2 = New TDTrackBar
                trackBar2.D2DParent = Me
                trackBar2.MinValue = 1
                trackBar2.MaxValue = 100
                trackBar2.Visible = False
                AddHandler trackBar2.Scroll, AddressOf OnTrackBar2Scroll
                AddHandler trackBar2.ValueChanged, Sub() label2.Text = Game.guiINI.getValue("creep_count").Replace("%count%", trackBar2.Value.ToString)

                upButton = New TDButton
                upButton.D2DParent = Me
                upButton.Text = Game.guiINI.getValue("wave_up")
                AddHandler upButton.Click, Sub()
                                               If listBox.SelectedIndex > 0 Then
                                                   Dim index = listBox.SelectedIndex
                                                   MoveItem(index, index - 1)
                                                   listBox.SelectedIndex = index - 1
                                               End If
                                           End Sub

                downButton = New TDButton
                downButton.D2DParent = Me
                downButton.Text = Game.guiINI.getValue("wave_down")
                AddHandler downButton.Click, Sub()
                                                 If listBox.SelectedIndex < spawn.Items.Count - 1 Then
                                                     Dim index = listBox.SelectedIndex
                                                     MoveItem(index, index + 1)
                                                     listBox.SelectedIndex = index + 1
                                                 End If
                                             End Sub

                newButton = New TDButton
                newButton.D2DParent = Me
                newButton.Text = Game.guiINI.getValue("wave_new")
                AddHandler newButton.Click, Sub() AddItem()

                removeButton = New TDButton
                removeButton.D2DParent = Me
                removeButton.Text = Game.guiINI.getValue("wave_remove")
                AddHandler removeButton.Click, Sub() RemoveItem(listBox.SelectedIndex)

                backButton = New TDButton
                backButton.D2DParent = Me
                backButton.Text = Game.guiINI.getValue("doneWE")
                AddHandler backButton.Click, Sub() goUp()

                AddHandler Me.Resize, Sub() ResizeControls()
                ResizeControls()
            End Sub

            Private Sub AddItem()
                Dim item As New AdvancedSpawnInfo.SpawnInfoItem
                item.Count = 1
                spawn.Items.Add(item)
                LoadList()
                listBox.SelectedIndex = spawn.Items.Count - 1
            End Sub

            Private Sub RemoveItem(index As Integer)
                If index < 0 Then Return
                spawn.Items.RemoveAt(index)
                LoadList()
            End Sub

            Private Sub MoveItem(index As Integer, destIndex As Integer)
                If index < 0 Then Return
                Dim item = spawn.Items(index)
                spawn.Items.RemoveAt(index)
                'If destIndex > index Then destIndex -= 1
                spawn.Items.Insert(destIndex, item)
                LoadList()
            End Sub

            Private Sub EditDialog_Show(sender As Object, e As EventArgs)
                If Index = waveDiag.spawnInf.Count Then
                    spawn = New AdvancedSpawnInfo
                    waveDiag.spawnInf.Add(spawn)
                Else
                    spawn = waveDiag.spawnInf(Index)
                End If
                textBox.Text = (spawn.Time / 60).ToString("0.0")
                LoadList()
            End Sub

            Private Sub OnSelectedIndexChanged(sender As Object, e As EventArgs)
                If listBox.SelectedIndex < 0 Then
                    trackBar1.Visible = False
                    trackBar2.Visible = False
                    label1.Visible = False
                    label2.Visible = False
                    textBox.Visible = True
                    label3.Visible = True
                Else
                    trackBar1.Visible = True
                    trackBar2.Visible = True
                    label1.Visible = True
                    label2.Visible = True
                    textBox.Visible = False
                    label3.Visible = False
                    Dim item = spawn.Items(listBox.SelectedIndex)
                    Dim val = -1
                    For i = 0 To creeps.Length - 1
                        If creeps(i).ID = item.Type Then
                            val = i
                            Exit For
                        End If
                    Next
                    trackBar1.Value = val
                    trackBar2.Value = item.Count
                End If
            End Sub

            Private Sub OnTrackBar1Scroll(sender As Object, e As EventArgs)
                If listBox.SelectedIndex >= 0 Then spawn.Items(listBox.SelectedIndex).Type = creeps(trackBar1.Value).ID
            End Sub

            Private Sub OnTrackBar2Scroll(sender As Object, e As EventArgs)
                If listBox.SelectedIndex >= 0 Then spawn.Items(listBox.SelectedIndex).Count = Convert.ToByte(trackBar2.Value)
            End Sub

            Private Sub LoadList()
                listBox.Items.Clear()
                For Each infoItem In spawn.Items
                    Dim item As New InfoListBoxitem
                    item.InfoItem = infoItem
                    listBox.Items.Add(item)
                Next
            End Sub

            Private Sub ResizeControls()
                Dim cWidth = Convert.ToInt32(Width * 0.32)
                Dim cBorderX = Convert.ToInt32(Width * 0.15)
                Dim cBorderY = Convert.ToInt32(Height * 0.1)
                listBox.Location = New Point(cBorderX, cBorderY)
                listBox.ItemHeight = Convert.ToInt32(25 * Game.getSizeFactor)
                Dim buttonSize As New Size(cWidth, Convert.ToInt32(50 * Game.getSizeFactor))
                Dim buttonBorder = Convert.ToInt32(15 * Game.getSizeFactor)
                listBox.Size = New Size(cWidth, 4 * buttonSize.Height + 3 * buttonBorder)
                upButton.Size = buttonSize
                downButton.Size = buttonSize
                newButton.Size = buttonSize
                removeButton.Size = buttonSize
                Dim buttonX = Width - cBorderX - cWidth
                upButton.Location = New Point(buttonX, cBorderY)
                downButton.Location = New Point(buttonX, cBorderY + buttonSize.Height + buttonBorder)
                newButton.Location = New Point(buttonX, cBorderY + 2 * (buttonSize.Height + buttonBorder))
                removeButton.Location = New Point(buttonX, cBorderY + 3 * (buttonSize.Height + buttonBorder))
                Dim trackBarWidth = Convert.ToInt32(600 * Game.getWidthFactor)
                Dim trackBarHeight = Convert.ToInt32(25 * Game.getHeightFactor)
                Dim trackBarSize As New Size(trackBarWidth, trackBarHeight)
                trackBar1.Size = trackBarSize
                trackBar2.Size = trackBarSize
                Dim trackBarX = (Width - trackBarWidth) \ 2
                trackBar1.Location = New Point(trackBarX, Convert.ToInt32(420 * getHeightFactor()))
                trackBar2.Location = New Point(trackBarX, Convert.ToInt32(500 * getHeightFactor()))
                backButton.Size = New Size(trackBarWidth, buttonSize.Height)
                backButton.Location = New Point(trackBarX, Convert.ToInt32(600 * getHeightFactor()))
                label1.Font = FontManager.GetFont("defaultLabelFormat")
                label2.Font = FontManager.GetFont("defaultLabelFormat")
                label3.Font = FontManager.GetFont("defaultLabelFormat")
                Dim labelHeight = Convert.ToInt32(40 * Game.getHeightFactor)
                Dim labelSize = New Size(trackBarWidth, labelHeight)
                label1.Size = labelSize
                label2.Size = labelSize
                label1.Location = New Point(trackBarX, trackBar1.Location.Y - labelHeight)
                label2.Location = New Point(trackBarX, trackBar2.Location.Y - labelHeight)
                textBox.Size = New Size(trackBarWidth, Convert.ToInt32(40 * Game.getHeightFactor))
                textBox.Location = New Point(trackBarX, Convert.ToInt32(460 * Game.getHeightFactor))
                label3.Size = labelSize
                label3.Location = New Point(trackBarX, textBox.Location.Y - labelHeight)
            End Sub

            Private Class InfoListBoxitem : Inherits TDListBox.ListBoxItem

                Friend Property InfoItem As AdvancedSpawnInfo.SpawnInfoItem

                Protected Overrides Sub DrawItem(rect As Rectangle, e As D2DDrawEventArgs)
                    If IsSelected Then e.Target.DrawRectangle(ColorManager.GetBrush("defaultWhite"), New RectangleF(rect.X + 0.5F, rect.Y + 0.5F, rect.Width - 1, rect.Height - 1))
                    If InfoItem Is Nothing Then Return
                    Dim countRect As New Rectangle(rect.X, rect.Y, Convert.ToInt32(50 * Game.getWidthFactor), rect.Height)
                    Dim formatName = "ListBoxTailingItemFormat"
                    If Not FontManager.FontRegistered(formatName) Then FontManager.RegisterFont(formatName, New FontTemplate("Polentical Neon", 16, TextAlignment:=TextAlignment.Trailing, ParagraphAlignment:=ParagraphAlignment.Center))
                    e.Target.DrawText(InfoItem.Count & "x", FontManager.GetFont(formatName), countRect, ColorManager.GetBrush("defaultWhite"))
                    Dim creepSize = Convert.ToInt32(rect.Height * 0.7)
                    Dim creepBorder = (rect.Height - creepSize) \ 2
                    Dim creepRect As New Rectangle(rect.X + countRect.Width + creepBorder, rect.Y + creepBorder, creepSize, creepSize)
                    CreepRenderer.DrawCreep(InfoItem.Type, e, creepRect)
                End Sub
            End Class
        End Class
    End Class

    Friend Class TowerDialog : Inherits Menu
        Private labels As List(Of D2DLabel)
        Private trackbars As List(Of TDTrackBar)
        Private backButton As TDButton
        Private topLabel As D2DLabel

        Friend Function GetAllowedTowers() As UInteger
            Return CalculateAllowedTowers()
        End Function

        Sub New(parent As WorldEditControl, initVal As UInteger)
            parentMenu = parent
            labels = New List(Of D2DLabel)
            trackbars = New List(Of TDTrackBar)

            Dim flags = BindingFlags.Static Or BindingFlags.NonPublic
            Dim towers() = Tower.GetAllTowers
            For i = 0 To towers.Length - 1
                Dim t = towers(i)
                Dim name = DirectCast(t.GetMethod("getName", flags).Invoke(Nothing, {}), String).Replace(Environment.NewLine, String.Empty)
                Dim l As New D2DLabel(Me)
                l.Text = name
                l.TextColor = DirectCast(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)
                l.Font = FontManager.GetFont("defaultLabelFormat")
                labels.Add(l)
                Dim tb As New TDTrackBar
                tb.D2DParent = Me
                tb.MinValue = 0
                tb.MaxValue = 1
                Dim towerInt = 1UI << i
                If (initVal And towerInt) = towerInt Then tb.Value = 1
                AddHandler tb.Scroll, Sub()
                                          If (DateTime.Now - lastMenuChange).Ticks < 100000 Then
                                              If (initVal And towerInt) = towerInt Then
                                                  tb.Value = 1
                                              Else
                                                  tb.Value = 0
                                              End If
                                          End If
                                      End Sub
                trackbars.Add(tb)
            Next

            backButton = New TDButton
            backButton.D2DParent = Me
            backButton.Text = Game.guiINI.getValue("back")
            AddHandler backButton.Click, Sub() goUp()

            topLabel = New D2DLabel(Me)
            topLabel.Text = Game.guiINI.getValue("towers_OnOff")
            topLabel.TextColor = DirectCast(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)
            topLabel.Font = FontManager.GetFont("defaultLabelFormat")

            AddHandler Me.Resize, Sub() ResizeControls()
            ResizeControls()
        End Sub

        Private Function CalculateAllowedTowers() As UInteger
            Dim allowedTowers = 0UI
            For i = 0 To 31
                Dim val = 1
                If trackbars.Count > i Then val = trackbars(i).Value
                Dim towerInt = Convert.ToUInt32(val) << i
                allowedTowers = allowedTowers Or towerInt
            Next
            Return allowedTowers
        End Function

        Private Sub ResizeControls()
            Dim tbHeight = 25 * Game.getHeightFactor
            Dim cWidth = 280 * Game.getWidthFactor
            Dim bWidth = Convert.ToInt32(600 * Game.getWidthFactor)
            Dim cX = (Width - bWidth) \ 2
            Dim cDistance = 25 * Game.getHeightFactor
            Dim completeHeight = trackbars.Count * tbHeight + (trackbars.Count - 1) * cDistance
            Dim cY = Convert.ToInt32((Height - completeHeight) / 2)
            Dim tbX = Convert.ToInt32(Width - cWidth - cX)
            Dim s = New Size(Convert.ToInt32(cWidth), Convert.ToInt32(tbHeight))
            For i = 0 To trackbars.Count - 1
                Dim y = cY + i * Convert.ToInt32(tbHeight + cDistance)
                labels(i).Location = New Point(cX, y)
                trackbars(i).Location = New Point(tbX, y)
                labels(i).Size = s
                trackbars(i).Size = s
                labels(i).Font = FontManager.GetFont("defaultLabelFormat")
            Next
            backButton.Size = New Size(bWidth, Convert.ToInt32(50 * Game.getHeightFactor))
            backButton.Location = New Point(cX, Convert.ToInt32(600 * Game.getHeightFactor))
            topLabel.TextColor = DirectCast(ColorManager.GetBrush("defaultWhite"), SolidColorBrush)
            topLabel.Font = FontManager.GetFont("defaultLabelFormat")
            topLabel.Size = s
            topLabel.Location = New Point(tbX, cY - Convert.ToInt32(30 * Game.getHeightFactor))
        End Sub
    End Class
End Class

Friend Class MapNotCustomException : Inherits Exception : End Class